package api

import (
	"net/http"
	"slices"
	"strconv"
	"time"

	m "0xacab.org/leap/menshen/pkg/models"
	"github.com/labstack/echo/v4"
)

// ListAllBridges godoc
// @Summary      Get All Bridges
// @Description  Fetch all bridges, optionally filtered by transport, port, obfuscation type and location. The returned list is also restricted by the BucketTokenAuth, an auth token which determines access to available resources. The order of bridges depends on the location of the client (which is provided by the cc query parameter) and should be followed by it.
// @Tags         Provisioning
// @Accept       json
// @Produce      json
// @Param        cc   query  string false "base country code (ISO-2)"
// @Param        tr	  query  string false "transport (tcp|udp)"
// @Param        port query  string false "port"
// @Param        type query  string false "type"
// @Param        loc  query  string false "location"
// @Success      200  {object}  []models.Bridge
// @Failure      400  {object}  error
// @Router       /api/5/bridges [get]
// @Security BucketTokenAuth
func (r *registry) ListAllBridges(c echo.Context) error {

	// Set Cache-Control header to instruct clients not to cache response
	c.Response().Header().Set("Cache-Control", "no-store")

	// Step 1: Validate parameters
	cc, err := sanitizeCountryCode(c.QueryParam(paramCountryCode))
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizeLocations(c.QueryParam("loc"), r.locations)
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizeTransport(c, []string{"udp", "tcp", "kcp", "quic"})
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizePort(c.QueryParam("port"))
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizeType(c, []string{"obfs4", "obfs4-hop"})
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	// Step 2: We start with all bridges
	bridges := r.AllBridges()

	// Step 3: apply gateway filters
	filters := bridgeFiltersFromParams(c, []string{"type", "loc", "port", "tr"})
	filters = maybeAddBridgeBucketFilter(c, filters)
	filters = maybeAddLastSeenBridgeCutoffFilter(r, filters)
	availableBridges := filter[*m.Bridge](alltrue(filters), bridges)

	// Step 4: sort gateways
	result, err := sortEndpoints(cc, availableBridges, r.locations, r.lm)
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	return c.JSON(http.StatusOK, result)
}

// I'd love for this to be generic across "endpoints" (ie bridges AND gateways)
// but golang generics/support of sum-types is still quite mediocre :/
// https://github.com/golang/go/issues/48522
// https://github.com/golang/go/issues/51259
// https://github.com/golang/go/issues/57644
func maybeAddBridgeBucketFilter(c echo.Context, filters []func(*m.Bridge) bool) []func(*m.Bridge) bool {
	buckets, ok := c.Get("buckets").([]string)
	if !ok {
		buckets = []string{""}
	}

	return append(filters, func(b *m.Bridge) bool {
		return slices.Contains(buckets, b.Bucket)
	})
}

// See comment on maybeAddBridgeBucketFilter for same issue re: making this generic across bridges and gateways
func maybeAddLastSeenBridgeCutoffFilter(r *registry, filters []func(*m.Bridge) bool) []func(*m.Bridge) bool {
	if r.lastSeenCutoffMillis == 0 {
		return filters
	}

	return append(filters, func(b *m.Bridge) bool {
		nowMillis := time.Now().UnixMilli()

		return b.LastSeenMillis >= nowMillis-r.lastSeenCutoffMillis
	})
}

func bridgeFiltersFromParams(c echo.Context, params []string) []func(*m.Bridge) bool {
	filters := make([]func(*m.Bridge) bool, 0)
	for _, param := range params {
		filters = maybeAddBridgeFilter(c, param, filters)
	}
	return filters
}

func maybeAddBridgeFilter(c echo.Context, param string, filters []func(*m.Bridge) bool) []func(*m.Bridge) bool {
	// get the query parameter
	q := c.QueryParam(param)
	if q == "" {
		// empty, return the array of filters
		return filters
	}

	var filter m.EndpointFilter[m.Bridge]
	switch param {
	case "type":
		filter = func(b *m.Bridge) bool {
			return b.Type == q
		}
	case "port":
		port, err := strconv.Atoi(q)
		if err != nil {
			return filters
		}
		filter = func(b *m.Bridge) bool {
			return b.Port == port
		}
	case "loc":
		filter = func(b *m.Bridge) bool {
			return b.Location == q
		}
	case "tr":
		filter = func(b *m.Bridge) bool {
			return b.Transport == q
		}
	default:
		return filters
	}
	return append(filters, filter)
}
