package api

import (
	"crypto/sha256"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"net/http"
	"net/http/httptest"
	"os"
	"strings"
	"testing"
	"time"

	m "0xacab.org/leap/menshen/pkg/models"
	"0xacab.org/leap/menshen/storage"
	"github.com/labstack/echo/v4"
	"github.com/stretchr/testify/assert"
)

func TestBridgeFilters(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	locationNY := "New York"
	locationMon := "Montreal"
	locationStruct := m.Location{
		CountryCode: "US",
		Label:       locationNY,
		DisplayName: locationNY,
		Lat:         "40.71",
		Lon:         "-74.00",
	}

	locationStructCA := m.Location{
		CountryCode: "CA",
		Label:       locationMon,
		DisplayName: locationMon,
		Lat:         "45.52",
		Lon:         "-73.65",
	}

	bridge1 := &m.Bridge{
		Host:      "bridge1",
		Port:      443,
		Transport: "TCP",
		Type:      "obfs4",
		Bucket:    "bucket1",
		Location:  locationNY,
	}
	bridge2 := &m.Bridge{
		Host:      "bridge2",
		Port:      443,
		Transport: "TCP",
		Type:      "obfs4",
		Location:  locationNY,
	}

	bridge3 := &m.Bridge{
		Host:           "bridge3",
		Port:           443,
		Transport:      "TCP",
		Type:           "obfs4",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Location:       locationNY,
	}

	bridge4 := &m.Bridge{
		Host:           "bridge4",
		Port:           443,
		Transport:      "TCP",
		Type:           "obfs4-hop",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Location:       locationNY,
	}

	bridge5 := &m.Bridge{
		Host:           "bridge5",
		Port:           443,
		Transport:      "quic",
		Type:           "obfs4-hop",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Location:       locationNY,
	}

	bridge6 := &m.Bridge{
		Host:           "bridge6",
		Port:           4430,
		Transport:      "kcp",
		Type:           "obfs4-hop",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Location:       locationMon,
	}

	testTable := []struct {
		name         string
		mockRegistry *registry
		parameter    string
		expected     func(*registry) string
		authToken    string
		dbAuthTokens []authTokenDbEntry
	}{
		{"no auth token only private bridges return empty",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct},
			},
			"?type=obfs4",
			func(r *registry) string {
				return "[]\n"
			},
			"",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token return one private bridge",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge2, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct},
			},
			"?type=obfs4",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge1})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token private and public bridges return private",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge2, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
			},
			"?type=obfs4",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge1})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token with lastSeenCutoffMillis enabled ignore offline bridge",
			&registry{
				bridges: bridgeMap{locationNY: []*m.Bridge{
					bridge1,
					bridge3,
				}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
				// Cut off is 5 seconds
				lastSeenCutoffMillis: 5000,
			},
			"?type=obfs4",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge3})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"transport quic return 1 bridge",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
			},
			"?tr=quic",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge5})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket2"}},
		},
		{"transport kcp return 1 bridge",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
			},
			"?tr=kcp",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge6})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket2"}},
		},
		{"location return 1 bridge",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
			},
			"?loc=Montreal",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge6})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket2"}},
		},
		{"port return 1 bridge",
			&registry{
				bridges:   bridgeMap{locationNY: []*m.Bridge{bridge1, bridge3, bridge4, bridge5}, locationMon: []*m.Bridge{bridge6}},
				locations: locationMap{locationNY: &locationStruct, locationMon: &locationStructCA},
			},
			"?port=4430",
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Bridge{bridge6})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket2"}},
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			expectedResponse := tc.expected(tc.mockRegistry)

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/bridges", tc.mockRegistry.ListAllBridges)

			// First test ListAllBridges
			req := httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/bridges%s", tc.parameter), nil)
			req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
			if tc.authToken != "" {
				req.Header.Set("x-menshen-auth-token", tc.authToken)
			}
			rec := httptest.NewRecorder()

			e.ServeHTTP(rec, req)
			assert.Equal(t, http.StatusOK, rec.Code)
			assert.Equal(t, strings.TrimSpace(expectedResponse), strings.TrimSpace(rec.Body.String()))
		})
	}
}

// if filtereing results in multiple Bridges being returned, we cannot do string comparisons in the
// test for verifying the result. This test has been moved (quick and dirty) from TestBridgeFilters.
func TestFiltersResultInRandomOrder(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	location1 := "New York"
	locationStruct := m.Location{
		CountryCode: "US",
		Label:       location1,
		DisplayName: location1,
		Lat:         "40.71",
		Lon:         "-74.00",
	}

	bridge1 := &m.Bridge{
		Host:      "bridge1",
		Port:      443,
		Transport: "TCP",
		Type:      "obfs4",
		Bucket:    "bucket1",
		Location:  location1,
	}

	bridge3 := &m.Bridge{
		Host:           "bridge3",
		Port:           443,
		Transport:      "TCP",
		Type:           "obfs4",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Location:       location1,
	}

	testTable := []struct {
		name         string
		mockRegistry *registry
		expected     []*m.Bridge
		authToken    string
		dbAuthTokens []authTokenDbEntry
	}{
		{"auth token with multiple buckets two private bridges",
			&registry{
				bridges:   bridgeMap{location1: []*m.Bridge{bridge1, bridge3}},
				locations: locationMap{location1: &locationStruct},
			},
			[]*m.Bridge{bridge1, bridge3},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/bridges", tc.mockRegistry.ListAllBridges)

			req := httptest.NewRequest(http.MethodGet, "/api/5/bridges?type=obfs4", nil)
			req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
			if tc.authToken != "" {
				req.Header.Set("x-menshen-auth-token", tc.authToken)
			}
			rec := httptest.NewRecorder()

			e.ServeHTTP(rec, req)
			assert.Equal(t, http.StatusOK, rec.Code)

			// Unmarshal the JSON data into the struct
			var bridges []*m.Bridge
			err = json.Unmarshal(rec.Body.Bytes(), &bridges)
			assert.Equal(t, err, nil)
			assert.ElementsMatch(t, tc.expected, bridges)
		})
	}

}
