// This file contains shared functionality between bridge and gateway endpoints
package api

import (
	"errors"
	"fmt"
	"slices"
	"strconv"
	"strings"

	"0xacab.org/leap/menshen/pkg/geolocate"
	"0xacab.org/leap/menshen/pkg/latency"
	m "0xacab.org/leap/menshen/pkg/models"
	wr "0xacab.org/leap/menshen/pkg/weightedrand"
	rbt "github.com/emirpasic/gods/v2/trees/redblacktree"
	"github.com/labstack/echo/v4"
)

func filter[T any](f func(t T) bool, items []T) []T {
	t := []T{}
	for _, item := range items {
		if f(item) {
			t = append(t, item)
		}
	}
	return t
}

func alltrue[T any](cond []func(T) bool) func(T) bool {
	return func(t T) bool {
		cur := true
		for _, f := range cond {
			cur = f(t) && cur
			if !cur {
				return cur
			}
		}
		return cur
	}
}

func sortEndpoints[T m.Bridge | m.Gateway](cc string, endpoints []*T, locations locationMap, lm *latency.Metric) ([]*T, error) {
	if len(endpoints) == 0 {
		return endpoints, nil
	}

	// without a country code we can only create a randomized list of endpoints
	if cc == "" {
		return wr.CreateWeightedRandomList(endpoints)
	}

	locationWeights, err := locationWeights(cc, locations, lm)
	if err != nil {
		return nil, err
	}

	tree := createLocationTree(endpoints, locationWeights)

	result := []*T{}
	// iterate over _lists_ of endpoints, grouped by location and sorted by distance
	for _, list := range tree.Values() {
		// add weighted randomized endpoints per location to result list
		result, err = wr.AppendWeightedRandomChoices(list, result)
		if err != nil {
			return nil, err
		}
	}

	return result, nil
}

func createLocationTree[T m.Bridge | m.Gateway](endpoints []*T, locationWeights map[string]float64) *rbt.Tree[float64, []wr.Choice[T]] {
	tree := rbt.New[float64, []wr.Choice[T]]()
	for _, endpoint := range endpoints {
		if endpoint == nil {
			continue
		}
		var list []wr.Choice[T]
		found := false
		locationWeight := locationWeights[getLocation(endpoint)]
		if list, found = tree.Get(locationWeight); !found {
			// TODO: we want to set weights per gateways, so that gateways with higher capacities
			// are priotized, the wr.Choice wrapper is a preparation for that
			list = []wr.Choice[T]{{Item: *endpoint, Weight: 1}}
		} else {
			list = append(list, wr.Choice[T]{Item: *endpoint, Weight: 1})
		}
		tree.Put(locationWeight, list)
	}
	return tree
}

func getLocation(endpoint interface{}) string {
	switch v := any(endpoint).(type) {
	case *m.Gateway:
		return v.Location
	case *m.Bridge:
		return v.Location
	default:
		return ""
	}
}

func locationWeights(cc string, locations locationMap, lm *latency.Metric) (map[string]float64, error) {
	locationWeights := make(map[string]float64)
	for _, location := range locations {
		lat, err := strconv.ParseFloat(location.Lat, 64)
		if err != nil {
			return nil, err
		}
		lon, err := strconv.ParseFloat(location.Lon, 64)
		if err != nil {
			return nil, err
		}

		point := &geolocate.Point{Lat: lat, Lon: lon}

		distance, err := geolocate.DistanceFromCountry(cc, point)
		if err != nil {
			return nil, err
		}

		millisecs := geolocate.GetEstimatedLatencyBetweenCountries(lm, cc, location.CountryCode)
		normalized := combine(normalizeDistance(distance), normalizeLatency(millisecs, lm.Max()), 2.0)
		// TODO: we might add here other factors such as regional preferences

		locationWeights[location.Label] = normalized
	}
	return locationWeights, nil
}

// normalizeDistance normalizes the haversine distance between two points to a value between 0 and 1
func normalizeDistance(distance float64) float64 {
	return (distance * 2) / geolocate.EARTH_CIRCUMFERENCE
}

// normalizeLatency normalizes the latency to a value between 0 and 1
func normalizeLatency(latency, maxLatency float64) float64 {
	return latency / maxLatency
}

// combine adds a new value to a base value and normalizes it to a value between 0 and 1
func combine(base, newVal, newValWeight float64) float64 {
	if newVal < 0 || newVal > 1 {
		// ignore newVal if it's not normalized
		return base
	}
	return (base + (newVal * newValWeight)) / (newValWeight + 1.0)
}

// sanitizeCountryCode will check if the passed ISO-2 country code is known to us.
// For empty strings, they stay the same.
// For non-valid country codes, it returns an error. Otherwise, it returns the original string.
func sanitizeCountryCode(cc string) (string, error) {
	if cc == "" {
		return "", nil
	}
	if len(cc) != 2 {
		return "", errors.New("unknown country")
	}
	cc = strings.ToUpper(cc)
	if !geolocate.IsValidCountryCode(cc) {
		return "", errors.New("unknown country")
	}
	return cc, nil
}

func sanitizeLocations(location string, locations locationMap) error {
	if location == "" {
		return nil
	}
	for k := range locations {
		if k == location {
			return nil
		}
	}
	return errors.New("unknown location")
}

func sanitizeTransport(c echo.Context, allowedTransports []string) error {
	return sanitizeToLower(c, "tr", allowedTransports)
}

func sanitizeType(c echo.Context, allowedTypes []string) error {
	return sanitizeToLower(c, "type", allowedTypes)
}

func sanitizeToLower(c echo.Context, paramKey string, allowedList []string) error {
	queryParam := c.QueryParam(paramKey)
	if queryParam == "" {
		return nil
	}
	queryParam = strings.ToLower(queryParam)
	if !slices.Contains(allowedList, queryParam) {
		return fmt.Errorf("unknown value %s for query param %s", c.QueryParam(paramKey), paramKey)
	}
	c.QueryParams().Set(paramKey, queryParam)
	return nil
}

func sanitizePort(port string) error {
	if port == "" {
		return nil
	}
	p, err := strconv.Atoi(port)
	if err != nil || p < 1 || p > 65535 {
		return errors.New("invalid port number")
	}
	return nil
}
