package api

import (
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"net/http/httptest"
	"testing"

	"0xacab.org/leap/menshen/pkg/latency"
	m "0xacab.org/leap/menshen/pkg/models"
	wr "0xacab.org/leap/menshen/pkg/weightedrand"
	"github.com/labstack/echo/v4"
	"github.com/magiconair/properties/assert"
	"github.com/stretchr/testify/require"
)

func TestSortEndpoints(t *testing.T) {
	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}

	locationStruct1 := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct2 := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct3 := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}
	gateway1 := &m.Gateway{Location: locations[0], Host: "gw1_ny"}
	gateway2 := &m.Gateway{Location: locations[0], Host: "gw2_ny"}
	gateway3 := &m.Gateway{Location: locations[1], Host: "gw3_par"}
	gateway4 := &m.Gateway{Location: locations[1], Host: "gw4_par"}
	gateway5 := &m.Gateway{Location: locations[2], Host: "gw5_mo"}
	gateway6 := &m.Gateway{Location: locations[2], Host: "gw6_mo"}

	lm, _ := latency.NewMetric()
	locationmap := locationMap{locations[0]: locationStruct1, locations[1]: locationStruct2, locations[2]: locationStruct3}
	gateways := []*m.Gateway{gateway1, gateway2, gateway3, gateway4, gateway5, gateway6}

	testTable := []struct {
		name        string
		cc          string
		expected    []string
		expextedErr error
	}{
		{"test 1 - client country code US - New York, Montreal, Paris",
			"US",
			[]string{"New York", "New York", "Montreal", "Montreal", "Paris", "Paris"},
			nil,
		},
		{"test 2 - client country code PL - Paris, New York, Montreal",
			"PL",
			[]string{"Paris", "Paris", "Montreal", "Montreal", "New York", "New York"},
			nil,
		},
		{"test 3 - client country code IS - Paris, New York, Montreal",
			"IS",
			[]string{"Paris", "Paris", "Montreal", "Montreal", "New York", "New York"},
			nil,
		},
		{"test 4 - client country code IN - Paris, New York, Montreal",
			"IN",
			[]string{"Paris", "Paris", "New York", "New York", "Montreal", "Montreal"},
			nil,
		},
		{"test 5 - client country code XX -  invalid",
			"XX",
			[]string{},
			fmt.Errorf("%v: %s", errors.New("unknown country code"), "XX"),
		},
	}

	for ti, test := range testTable {
		t.Run(test.name, func(t *testing.T) {
			sortedList, err := sortEndpoints(test.cc, gateways, locationmap, lm)
			assert.Equal(t, err, test.expextedErr)
			assert.Equal(t, len(sortedList), len(test.expected))
			for i, entry := range sortedList {
				assert.Equal(t, entry.Location, test.expected[i], fmt.Sprintf("test %d: item %d", ti+1, i+1))
			}
		})
	}
}

func TestSortGateways_RandomOrderWithinLocation(t *testing.T) {
	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}

	locationStruct1 := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct2 := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct3 := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}
	gateway1 := &m.Gateway{Location: locations[0], Host: "gw1_ny"}
	gateway2 := &m.Gateway{Location: locations[0], Host: "gw2_ny"}
	gateway3 := &m.Gateway{Location: locations[1], Host: "gw3_par"}
	gateway4 := &m.Gateway{Location: locations[1], Host: "gw4_par"}
	gateway5 := &m.Gateway{Location: locations[2], Host: "gw5_mo"}
	gateway6 := &m.Gateway{Location: locations[2], Host: "gw6_mo"}

	lm, _ := latency.NewMetric()
	locationmap := locationMap{locations[0]: locationStruct1, locations[1]: locationStruct2, locations[2]: locationStruct3}
	gateways := []*m.Gateway{gateway1, gateway2, gateway3, gateway4, gateway5, gateway6}

	firstResult, _ := sortEndpoints("US", gateways, locationmap, lm)
	differentHosts := map[string]bool{"New York": false, "Paris": false, "Montreal": false}
	differentOrder := false
	for range 100 {
		sortedList, _ := sortEndpoints("US", gateways, locationmap, lm)
		assert.Equal(t, sortedList[0].Location, "New York")
		assert.Equal(t, sortedList[2].Location, "Montreal")
		assert.Equal(t, sortedList[4].Location, "Paris")
		if sortedList[0].Host != firstResult[0].Host {
			differentHosts[sortedList[0].Location] = true
		}
		if sortedList[2].Host != firstResult[2].Host {
			differentHosts[sortedList[2].Location] = true
		}
		if sortedList[4].Host != firstResult[4].Host {
			differentHosts[sortedList[4].Location] = true
		}
		if differentHosts["New York"] && differentHosts["Paris"] && differentHosts["Montreal"] {
			differentOrder = true
			break
		}
	}
	assert.Equal(t, differentOrder, true)
}

func TestSortBridges_RandomOrderWithinLocation(t *testing.T) {
	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}

	locationStruct1 := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct2 := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct3 := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}

	bridge1 := &m.Bridge{Location: locations[0], Host: "gw1_ny"}
	bridge2 := &m.Bridge{Location: locations[0], Host: "gw2_ny"}
	bridge3 := &m.Bridge{Location: locations[1], Host: "gw3_par"}
	bridge4 := &m.Bridge{Location: locations[1], Host: "gw4_par"}
	bridge5 := &m.Bridge{Location: locations[2], Host: "gw5_mo"}
	bridge6 := &m.Bridge{Location: locations[2], Host: "gw6_mo"}

	lm, _ := latency.NewMetric()
	locationmap := locationMap{locations[0]: locationStruct1, locations[1]: locationStruct2, locations[2]: locationStruct3}
	bridges := []*m.Bridge{bridge1, bridge2, bridge3, bridge4, bridge5, bridge6}

	firstResult, _ := sortEndpoints("US", bridges, locationmap, lm)
	differentHosts := map[string]bool{"New York": false, "Paris": false, "Montreal": false}
	differentOrder := false
	for range 100 {
		sortedList, _ := sortEndpoints("US", bridges, locationmap, lm)
		assert.Equal(t, sortedList[0].Location, "New York")
		assert.Equal(t, sortedList[2].Location, "Montreal")
		assert.Equal(t, sortedList[4].Location, "Paris")
		if sortedList[0].Host != firstResult[0].Host {
			differentHosts[sortedList[0].Location] = true
		}
		if sortedList[2].Host != firstResult[2].Host {
			differentHosts[sortedList[2].Location] = true
		}
		if sortedList[4].Host != firstResult[4].Host {
			differentHosts[sortedList[4].Location] = true
		}
		if differentHosts["New York"] && differentHosts["Paris"] && differentHosts["Montreal"] {
			differentOrder = true
			break
		}
	}
	assert.Equal(t, differentOrder, true)
}

func TestSanitizeCountryCode(t *testing.T) {
	inputs := []string{"invalid", "XX", "123"}
	for _, input := range inputs {
		_, err := sanitizeCountryCode(input)
		assert.Equal(t, err, errors.New("unknown country"))
	}

	inputs = []string{"SV", "us", "uS", ""}
	expected := []string{"SV", "US", "US", ""}
	for i, input := range inputs {
		sanitized, err := sanitizeCountryCode(input)
		assert.Equal(t, err, nil)
		assert.Equal(t, sanitized, expected[i])
	}
}

func TestSanitizePort(t *testing.T) {
	assert.Equal(t, sanitizePort("invalid"), errors.New("invalid port number"))
	assert.Equal(t, sanitizePort("-1"), errors.New("invalid port number"))
	assert.Equal(t, sanitizePort("100000000"), errors.New("invalid port number"))
	assert.Equal(t, sanitizePort("443"), nil)
	assert.Equal(t, sanitizePort(""), nil)
}

func TestSanitizeTransport(t *testing.T) {
	testTable := []struct {
		name              string
		url               string
		allowedTransports []string
		expectedParam     string
		expextedErr       error
	}{
		{
			"transport allowed",
			"/dummyendpoint?tr=udp",
			[]string{"tcp", "udp"},
			"udp",
			nil,
		},
		{
			"transport not allowed return error",
			"/dummyendpoint?tr=udp",
			[]string{"tcp"},
			"udp",
			errors.New("unknown value udp for query param tr"),
		},
		{
			"no transport parameter",
			"/dummyendpoint",
			[]string{"tcp", "udp"},
			"",
			nil,
		},
		{
			"uppercase transport parameter to lowercase",
			"/dummyendpoint?tr=UDP",
			[]string{"tcp", "udp"},
			"udp",
			nil,
		},
	}
	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {
			e := echo.New()
			req := httptest.NewRequest(http.MethodGet, tc.url, nil)
			c := e.NewContext(req, nil)
			assert.Equal(t, sanitizeTransport(c, tc.allowedTransports), tc.expextedErr)
			assert.Equal(t, c.QueryParam("tr"), tc.expectedParam)
		})
	}
}

func TestSanitizeType(t *testing.T) {
	testTable := []struct {
		name              string
		url               string
		allowedTransports []string
		expectedParam     string
		expextedErr       error
	}{
		{
			"transport allowed",
			"/dummyendpoint?type=obfs4",
			[]string{"obfs4-hop", "obfs4"},
			"obfs4",
			nil,
		},
		{
			"transport not allowed return error",
			"/dummyendpoint?type=kcp",
			[]string{"obfs4-hop", "obfs4"},
			"kcp",
			errors.New("unknown value kcp for query param type"),
		},
		{
			"no transport parameter",
			"/dummyendpoint",
			[]string{"obfs4-hop", "obfs4"},
			"",
			nil,
		},
		{
			"uppercase transport parameter to lowercase",
			"/dummyendpoint?type=OBFS4-HOP",
			[]string{"obfs4-hop", "obfs4"},
			"obfs4-hop",
			nil,
		},
	}
	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {
			e := echo.New()
			req := httptest.NewRequest(http.MethodGet, tc.url, nil)
			c := e.NewContext(req, nil)
			assert.Equal(t, sanitizeType(c, tc.allowedTransports), tc.expextedErr)
			assert.Equal(t, c.QueryParam("type"), tc.expectedParam)
		})
	}
}

func TestSanitizeLocation(t *testing.T) {
	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}

	locationStruct1 := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct2 := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct3 := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}

	locationmap := locationMap{locations[0]: locationStruct1, locations[1]: locationStruct2, locations[2]: locationStruct3}

	assert.Equal(t, sanitizeLocations("Moscow", locationmap), errors.New("unknown location"))
	assert.Equal(t, sanitizeLocations("Paris", locationmap), nil)
	assert.Equal(t, sanitizeLocations("", locationmap), nil)
}

func TestCreateWeightedRandomList(t *testing.T) {
	bridge1 := &m.Bridge{Location: "ny", Host: "gw1_ny"}
	bridge2 := &m.Bridge{Location: "ny", Host: "gw2_ny"}
	list := []*m.Bridge{bridge1, bridge2}

	bridge1First := false
	bridge2First := false
	for range 100 {
		randomlist, err := wr.CreateWeightedRandomList(list)
		assert.Equal(t, nil, err)
		bridge1Found := false
		bridge2Found := false
		for i, element := range randomlist {
			// "deep" comparison of pointers content via marshalling work-around
			jsonElement, err := toJson(*element)
			require.NoError(t, err)
			jsonBridge1, err := toJson(*bridge1)
			require.NoError(t, err)
			jsonBridge2, err := toJson(*bridge2)
			require.NoError(t, err)

			if jsonElement == jsonBridge1 {
				bridge1Found = true
				if i == 0 {
					bridge1First = true
				}
			} else if jsonElement == jsonBridge2 {
				bridge2Found = true
				if i == 0 {
					bridge2First = true
				}
			}
		}
		assert.Equal(t, bridge1Found, true, "bridge 1 found in randomized list")
		assert.Equal(t, bridge2Found, true, "bridge 2 found in randomized list")
	}
	assert.Equal(t, bridge1First, true, "bridge 1 at least once at first position in randomized lsit")
	assert.Equal(t, bridge2First, true, "bridge 2 at least once at first position in randomized lsit")

}

func toJson(model any) (string, error) {
	res, err := json.Marshal(model)
	if err != nil {
		return "", err
	}
	return string(res), nil
}
