package api

import (
	"net/http"
	"strconv"
	"time"

	"github.com/labstack/echo/v4"

	"slices"

	m "0xacab.org/leap/menshen/pkg/models"
)

var (
	paramCountryCode     = "cc"
	paramsGetAllGateways = []string{"tr", "port", "loc"}
)

// ListAllGateways godoc
// @Summary      Get All Gateways
// @Description  Fetch all gateways, optionally filtered by transport, port and location. The returned list is also restricted by the BucketTokenAuth, an auth token which determines access to available resources. The order of gateways depends on the location of the client (which is provided by the cc query parameter) and should be followed by it.
// @Tags         Provisioning
// @Accept       json
// @Produce      json
// @Param        cc   query  string false "base country code (ISO-2)"
// @Param        tr	  query  string false "transport (tcp|udp)"
// @Param        port query  string false "port"
// @Param        loc  query  string false "location override"
// @Success      200  {object}  []models.Gateway
// @Failure      400  {object}  error
// @Router       /api/5/gateways [get]
// @Security BucketTokenAuth
func (r *registry) ListAllGateways(c echo.Context) error {

	// Set Cache-Control header to instruct clients not to cache response
	c.Response().Header().Set("Cache-Control", "no-store")

	// Step 1: Validate parameters
	cc, err := sanitizeCountryCode(c.QueryParam(paramCountryCode))
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizeLocations(c.QueryParam("loc"), r.locations)
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizeTransport(c, []string{"udp", "tcp"})
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	err = sanitizePort(c.QueryParam("port"))
	if err != nil {
		return c.JSON(http.StatusBadRequest, err)
	}

	doIncrementCountryHit(cc)

	// Step 2: We start with all the gateways
	gateways := r.AllGateways()

	// Step 3: apply gateway filters
	// TODO: we may want to execute filtering after each step here, and return a specific error
	// that the client can interpret (or at least that helps debugging)
	gwFilters := gatewayFiltersFromParams(c, paramsGetAllGateways)
	gwFilters = maybeAddGatewayBucketFilter(c, gwFilters)
	gwFilters = maybeAddLastSeenGatewayCutoffFilter(r, gwFilters)
	availableGateways := filter[*m.Gateway](alltrue(gwFilters), gateways)

	// Step 4: sort gateways
	result, err := sortEndpoints(cc, availableGateways, r.locations, r.lm)
	if err != nil {
		return c.JSON(http.StatusBadRequest, nil)
	}

	return c.JSON(http.StatusOK, result)
}

func gatewayFiltersFromParams(c echo.Context, params []string) []func(*m.Gateway) bool {
	filters := make([]func(*m.Gateway) bool, 0)
	for _, param := range params {
		filters = maybeAddGatewayFilter(c, param, filters)
	}
	return filters
}

// maybeAddGatewayFilter receives a context object, a parameter string, and an array of filters. If the query parameter is found, it expands the filter array by adding the matching gateway filter.
func maybeAddGatewayFilter(c echo.Context, param string, filters []func(*m.Gateway) bool) []func(*m.Gateway) bool {
	// get the query parameter
	q := c.QueryParam(param)
	if q == "" {
		// empty, return the array of filters
		return filters
	}

	var filter m.EndpointFilter[m.Gateway]
	switch param {
	case "tr":
		filter = func(gw *m.Gateway) bool {
			return gw.Transport == q
		}
	case "port":
		// TODO: we could also consider a special comma-separated "ports". keeeping it simple for now
		port, err := strconv.Atoi(q)
		if err != nil {
			return filters
		}
		filter = func(gw *m.Gateway) bool {
			return gw.Port == port
		}
	case "loc":
		filter = func(gw *m.Gateway) bool {
			return gw.Location == q
		}

	default:
		// not a known filter
		return filters
	}
	return append(filters, filter)
}

func maybeAddLastSeenGatewayCutoffFilter(r *registry, filters []func(*m.Gateway) bool) []func(*m.Gateway) bool {
	if r.lastSeenCutoffMillis == 0 {
		return filters
	}

	return append(filters, func(b *m.Gateway) bool {
		nowMillis := time.Now().UnixMilli()

		return b.LastSeenMillis >= nowMillis-r.lastSeenCutoffMillis
	})
}

func maybeAddGatewayBucketFilter(c echo.Context, filters []func(*m.Gateway) bool) []func(*m.Gateway) bool {
	buckets, ok := c.Get("buckets").([]string)
	if !ok {
		buckets = []string{""}
	}

	return append(filters, func(b *m.Gateway) bool {
		return slices.Contains(buckets, b.Bucket)
	})
}
