package api

import (
	"crypto/sha256"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"net/http"
	"net/http/httptest"
	"net/url"
	"os"
	"strings"
	"testing"
	"time"

	"0xacab.org/leap/menshen/pkg/latency"
	m "0xacab.org/leap/menshen/pkg/models"
	"0xacab.org/leap/menshen/storage"
	"github.com/labstack/echo/v4"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestBucketAndCutOffFiltersFromCC_US(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	lm, err := latency.NewMetric()
	require.Nil(t, err)

	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}
	locationStruct_NY := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct_Par := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct_Mon := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}
	locationsMap := locationMap{locations[0]: locationStruct_NY, locations[1]: locationStruct_Par, locations[2]: locationStruct_Mon}

	gateway1_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw1_NY",
	}

	gateway2_Par := &m.Gateway{
		Location: locations[1],
		Type:     "openvpn",
		Host:     "gw2_Par",
	}

	gateway3_Mon := &m.Gateway{
		Location: locations[2],
		Type:     "openvpn",
		Bucket:   "bucket2",
		Host:     "gw3_Mon",
	}

	gateway4_Par := &m.Gateway{
		Location: locations[1],
		Type:     "openvpn",
		Bucket:   "bucket2",
		Host:     "gw4_Par",
	}

	gateway5_NY_offline := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli() - int64(time.Second*6),
		Host:           "gw5_NY",
	}

	testTable := []struct {
		name         string
		mockRegistry *registry
		expected     func(*registry) string
		authToken    string
		dbAuthTokens []authTokenDbEntry
	}{
		{"no auth token only private gateways return empty",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			func(r *registry) string {
				return "[]\n"
			},
			"",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"no auth token public and private gateways return public",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par, gateway4_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Gateway{gateway2_Par})
				assert.NoError(t, err)
				return string(bytes)
			},
			"",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token one private gateway return 1 private gateway",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			func(r *registry) string {
				bytes, err := json.Marshal(r.gateways[locations[0]])
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token one private one public gateway return 1 private gateway",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			func(r *registry) string {
				bytes, err := json.Marshal(r.gateways[locations[0]])
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1"}},
		},
		{"auth token with multiple buckets return three private gateways",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par, gateway4_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Gateway{gateway1_NY, gateway3_Mon, gateway4_Par})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"auth token with lastSeenCutoffMillis enabled ignore offline gateway",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway5_NY_offline}, locations[1]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
				// Cutoff of 5 seconds
				lastSeenCutoffMillis: 5000,
			},
			func(r *registry) string {
				bytes, err := json.Marshal([]*m.Gateway{gateway1_NY})
				assert.NoError(t, err)
				return string(bytes)
			},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			expectedResponse := tc.expected(tc.mockRegistry)

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/gateways", tc.mockRegistry.ListAllGateways)

			// First test ListAllGateways
			req := httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/gateways?cc=%v", url.QueryEscape("US")), nil)
			req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
			if tc.authToken != "" {
				req.Header.Set("x-menshen-auth-token", tc.authToken)
			}
			rec := httptest.NewRecorder()

			e.ServeHTTP(rec, req)
			assert.Equal(t, http.StatusOK, rec.Code)
			assert.Equal(t, strings.TrimSpace(expectedResponse), strings.TrimSpace(rec.Body.String()))
		})
	}
}

func TestGatewayOrderingForDifferentCCs(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	lm, err := latency.NewMetric()
	require.Nil(t, err)

	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}
	locationStruct_NY := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct_Par := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct_Mon := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}
	locationsMap := locationMap{locations[0]: locationStruct_NY, locations[1]: locationStruct_Par, locations[2]: locationStruct_Mon}

	gateway1_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw1_NY",
	}

	gateway2_Par := &m.Gateway{
		Location: locations[1],
		Type:     "openvpn",
		Bucket:   "bucket1",
		Host:     "gw2_Par",
	}

	gateway3_Mon := &m.Gateway{
		Location: locations[2],
		Type:     "openvpn",
		Bucket:   "bucket2",
		Host:     "gw3_Mon",
	}

	testTable := []struct {
		name                    string
		mockRegistry            *registry
		countryCode             string
		expectedOrderedGateways []*m.Gateway
		authToken               string
		dbAuthTokens            []authTokenDbEntry
	}{
		{"ordering CC US",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"US",
			[]*m.Gateway{gateway1_NY, gateway3_Mon, gateway2_Par},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"ordering CC France",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"FR",
			[]*m.Gateway{gateway2_Par, gateway3_Mon, gateway1_NY},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"ordering CC Canada",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"CA",
			[]*m.Gateway{gateway3_Mon, gateway1_NY, gateway2_Par},
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			expected := func() string {
				bytes, err := json.Marshal(tc.expectedOrderedGateways)
				assert.NoError(t, err)
				return string(bytes)
			}

			expectedResponse := expected()

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/gateways", tc.mockRegistry.ListAllGateways)

			req := httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/gateways?cc=%v", url.QueryEscape(tc.countryCode)), nil)
			req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
			if tc.authToken != "" {
				req.Header.Set("x-menshen-auth-token", tc.authToken)
			}
			rec := httptest.NewRecorder()

			e.ServeHTTP(rec, req)
			assert.Equal(t, http.StatusOK, rec.Code)
			assert.Equal(t, strings.TrimSpace(expectedResponse), strings.TrimSpace(rec.Body.String()))
		})
	}
}

func TestLocationFilters(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	lm, err := latency.NewMetric()
	require.Nil(t, err)

	locations := []string{"New York", "Paris", "Montreal"}
	countryCodes := []string{"US", "FR", "CA"}
	locationStruct_NY := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct_Par := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationStruct_Mon := &m.Location{
		CountryCode: countryCodes[2],
		DisplayName: locations[2],
		Label:       locations[2],
		Lat:         "45.50",
		Lon:         "-73.56",
	}
	locationsMap := locationMap{locations[0]: locationStruct_NY, locations[1]: locationStruct_Par, locations[2]: locationStruct_Mon}

	gateway1_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw1_NY",
	}

	gateway2_Par := &m.Gateway{
		Location: locations[1],
		Type:     "openvpn",
		Bucket:   "bucket1",
		Host:     "gw2_Par",
	}

	gateway3_Mon := &m.Gateway{
		Location: locations[2],
		Type:     "openvpn",
		Bucket:   "bucket2",
		Host:     "gw3_Mon",
	}

	testTable := []struct {
		name         string
		mockRegistry *registry
		countryCode  string
		authToken    string
		dbAuthTokens []authTokenDbEntry
	}{
		{"Location filters CC US",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"US",
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"Location filters CC France",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"FR",
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
		{"Location filters CC Canada",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY}, locations[1]: []*m.Gateway{gateway2_Par}, locations[2]: []*m.Gateway{gateway3_Mon}},
				locations: locationsMap,
				lm:        lm,
			},
			"CA",
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/gateways", tc.mockRegistry.ListAllGateways)

			for _, loc := range locations {
				expected := func(r *registry) string {
					bytes, err := json.Marshal(r.gateways[loc])
					assert.NoError(t, err)
					return string(bytes)
				}

				expectedResponse := expected(tc.mockRegistry)

				// test location parameter
				req := httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/gateways?loc=%v", url.QueryEscape(loc)), nil)
				req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
				if tc.authToken != "" {
					req.Header.Set("x-menshen-auth-token", tc.authToken)
				}
				rec := httptest.NewRecorder()

				e.ServeHTTP(rec, req)
				assert.Equal(t, http.StatusOK, rec.Code)
				assert.Equal(t, strings.TrimSpace(expectedResponse), strings.TrimSpace(rec.Body.String()))

				// verify that location parameter takes precendence over country code
				req = httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/gateways?cc=%v&loc=%v", url.QueryEscape(tc.countryCode), url.QueryEscape(loc)), nil)
				req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
				if tc.authToken != "" {
					req.Header.Set("x-menshen-auth-token", tc.authToken)
				}
				rec = httptest.NewRecorder()

				e.ServeHTTP(rec, req)
				assert.Equal(t, http.StatusOK, rec.Code)
				assert.Equal(t, strings.TrimSpace(expectedResponse), strings.TrimSpace(rec.Body.String()))
			}
		})
	}
}

func TestTransportAndPortFilters(t *testing.T) {
	type authTokenDbEntry struct {
		key     string
		buckets string
	}

	lm, err := latency.NewMetric()
	require.Nil(t, err)

	locations := []string{"New York", "Paris"}
	countryCodes := []string{"US", "FR"}
	locationStruct_NY := &m.Location{
		CountryCode: countryCodes[0],
		DisplayName: locations[0],
		Label:       locations[0],
		Lat:         "40.71",
		Lon:         "-74.00",
	}
	locationStruct_Par := &m.Location{
		CountryCode: countryCodes[1],
		DisplayName: locations[1],
		Label:       locations[1],
		Lat:         "46.22",
		Lon:         "2.21",
	}
	locationsMap := locationMap{locations[0]: locationStruct_NY, locations[1]: locationStruct_Par}

	gateway1_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw1_NY",
		Port:           1194,
		Transport:      "tcp",
	}

	gateway2_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw2_NY",
		Port:           1194,
		Transport:      "udp",
	}

	gateway3_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw3_NY",
		Port:           443,
		Transport:      "tcp",
	}
	gateway4_NY := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket2",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw4_NY",
		Port:           443,
		Transport:      "tcp",
	}

	gateway5_Par := &m.Gateway{
		Location:       locations[0],
		Type:           "openvpn",
		Bucket:         "bucket1",
		LastSeenMillis: time.Now().UnixMilli(),
		Host:           "gw5_Par",
		Port:           443,
		Transport:      "tcp",
	}

	type empty struct{}

	testTable := []struct {
		name                 string
		mockRegistry         *registry
		countryCode          string
		transport            string
		port                 int
		authToken            string
		dbAuthTokens         []authTokenDbEntry
		expectedGateways     interface{}
		expectedResponseCode int
	}{
		{"transport tcp port 443 return 3 GWs",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"US",
			"tcp",
			443,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			[]*m.Gateway{gateway3_NY, gateway4_NY, gateway5_Par},
			200,
		},
		{"transport tcp port 1194 return 1 gw",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"FR",
			"tcp",
			1194,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			[]*m.Gateway{gateway1_NY},
			200,
		},
		{"transport tcp port 80 return empty",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"CA",
			"tcp",
			80,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			[]*m.Gateway{},
			200,
		},
		{"transport invalid return empty",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"US",
			"invalid",
			443,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			empty{},
			400,
		},
		{"port invalid return empty",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"FR",
			"udp",
			0,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			empty{},
			400,
		},
		{"transport udp port 1194 return 1 GW",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"CA",
			"udp",
			1194,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket1,bucket2"}},
			[]*m.Gateway{gateway2_NY},
			200,
		},
		{"transport tcp port 443 bucket 2 only return 1 GW",
			&registry{
				gateways:  gatewayMap{locations[0]: []*m.Gateway{gateway1_NY, gateway2_NY, gateway3_NY, gateway4_NY}, locations[1]: []*m.Gateway{gateway5_Par}},
				locations: locationsMap,
				lm:        lm,
			},
			"US",
			"tcp",
			443,
			"key1",
			[]authTokenDbEntry{{"key1", "bucket2"}},
			[]*m.Gateway{gateway4_NY},
			200,
		},
	}

	for _, tc := range testTable {
		t.Run(tc.name, func(t *testing.T) {

			dir, err := os.MkdirTemp("", "")
			assert.NoError(t, err)
			defer os.RemoveAll(dir)

			db, err := storage.OpenDatabase(dir + "/db1.sql")
			assert.NoError(t, err)
			defer db.Close()

			stmt, err := db.Prepare("INSERT INTO tokens(key, buckets) VALUES(?, ?)")
			assert.NoError(t, err)

			for _, token := range tc.dbAuthTokens {
				h := sha256.New()
				h.Write([]byte(token.key))
				authTokenHashBytes := h.Sum(nil)

				authTokenHashString := base64.StdEncoding.EncodeToString(authTokenHashBytes)
				_, err = stmt.Exec(authTokenHashString, token.buckets)
				assert.NoError(t, err)
			}

			e := echo.New()
			e.Use(storageMiddleware(db))
			e.Use(authTokenMiddleware)
			e.GET("/api/5/gateways", tc.mockRegistry.ListAllGateways)

			// test cc, transport and port parameters along with buckets
			req := httptest.NewRequest(http.MethodGet, fmt.Sprintf("/api/5/gateways?cc=%v&tr=%v&port=%d", url.QueryEscape(tc.countryCode), url.QueryEscape(tc.transport), tc.port), nil)
			req.Header.Set(echo.HeaderContentType, echo.MIMEApplicationJSON)
			if tc.authToken != "" {
				req.Header.Set("x-menshen-auth-token", tc.authToken)
			}
			rec := httptest.NewRecorder()

			e.ServeHTTP(rec, req)
			assert.Equal(t, tc.expectedResponseCode, rec.Code)

			var responseStruct []*m.Gateway
			responseString := strings.TrimSpace(rec.Body.String())
			err = json.Unmarshal([]byte(responseString), &responseStruct)
			if err == nil {
				assert.Equal(t, len(tc.expectedGateways.([]*m.Gateway)), len(responseStruct))
				for _, gw := range responseStruct {
					assert.Contains(t, tc.expectedGateways, gw)
				}
			} else {
				// unmarshalling the response body to an gateway slice is only allowed to fail in case we get a http error code
				assert.Equal(t, 400, rec.Code)
			}
		})
	}
}
