// Package weightedrand contains a data structure and algorithm used
// to randomly select an element from some kind of list, where the chances of
// each element to be selected not being equal, but defined by relative
// "weights" (or probabilities). This is called weighted random selection.
//
// API inspied by mroth's https://github.com/mroth/weightedrand/
// adapted to avoid duplicate picks from the source list and opptimized for
// smaller sets (up to 1e4)

package weightedrand

import (
	"fmt"
	"math/rand"
	"sort"

	"errors"
)

// Choice is a generic wrapper that can be used to add weights for any item.
type Choice[T any] struct {
	Item   T
	Weight int
}

// A Chooser caches many possible Choices in a structure designed to improve
// performance on repeated calls for weighted random selection.
type Chooser[T any] struct {
	items       []Choice[T]
	totalWeight int
}

// NewChooser initializes a new Chooser for picking from the provided choices.
func NewChooser[T any](choices ...Choice[T]) (*Chooser[T], error) {
	sort.Slice(choices, func(i, j int) bool {
		return choices[i].Weight < choices[j].Weight
	})
	chooser := Chooser[T]{items: choices}
	chooser.totalWeight = 0
	for _, item := range chooser.items {
		if item.Weight < 1 {
			return nil, errors.New("item weight cannot be < 1")
		}
		chooser.totalWeight += int(item.Weight)
	}
	return &chooser, nil
}

// PickRandom removes a random item from the list based on weights and returns it.
// This method is not thread safe.
func (c *Chooser[T]) PickRandom() (*T, error) {
	if len(c.items) == 0 {
		return nil, fmt.Errorf("no items left to pick")
	}
	if len(c.items) == 1 {
		return &c.items[0].Item, nil
	}

	// Generate a random number between 0 and totalWeight
	r := rand.Intn(c.totalWeight)

	// Find the item corresponding to the random number
	for i, item := range c.items {
		if r < item.Weight {
			// Remove the item from the list, that's ok
			// within the loop since we're returning here
			chosenItem := c.items[i]
			c.items = append(c.items[:i], c.items[i+1:]...)
			// keep total weight in sync with sum of item weights
			c.totalWeight -= chosenItem.Weight
			return &chosenItem.Item, nil
		}
		r -= item.Weight
	}

	return nil, fmt.Errorf("failed to pick an item")
}

// CreateWeightedRandomList creates weighted randomized slice containing all elements of the source slice
func CreateWeightedRandomList[T any](source []*T) ([]*T, error) {
	list := []Choice[T]{}
	for _, element := range source {
		if element == nil {
			continue
		}
		list = append(list, Choice[T]{Item: *element, Weight: 1})
	}
	result := []*T{}
	return AppendWeightedRandomChoices(list, result)
}

// AppendWeightedRandomChoices appends weighted randomized elements to a slice.
// The order of previously added elements in the result list is retained
func AppendWeightedRandomChoices[T any](choices []Choice[T], result []*T) ([]*T, error) {
	chooser, err := NewChooser(
		choices...,
	)
	if err != nil {
		return nil, err
	}
	for range choices {
		element, err := chooser.PickRandom()
		if err != nil {
			return nil, fmt.Errorf("failed to create weighted random list: %v", err)
		}
		result = append(result, element)
	}
	return result, nil
}
