package weightedrand

import (
	"fmt"
	"math"
	"math/rand"
	"testing"

	"github.com/jmcvetta/randutil"
	mrothwr "github.com/mroth/weightedrand/v2"
	"github.com/tj/assert"
)

const BMMinChoices = 10
const BMMaxChoices = 10_000_000

func BenchmarkMultiple(b *testing.B) {
	for n := BMMinChoices; n <= BMMaxChoices; n *= 10 {
		b.Run(fmt.Sprintf("size=%s", fmt1eN(n)), func(b *testing.B) {
			wr_choices := mockChoices(b, n)
			ru_choices := convertChoices(b, wr_choices)
			wr_menshen_choices := convertChoicesToMenshenWeightedRand(b, wr_choices)

			b.Run("concurrency=single", func(b *testing.B) {
				b.Run("lib=randutil", func(b *testing.B) {
					for i := 0; i < b.N; i++ {
						_, _ = randutil.WeightedChoice(ru_choices)
					}
				})

				b.Run("lib=mrothwr weightedrand", func(b *testing.B) {
					chs, err := mrothwr.NewChooser(wr_choices...)
					if err != nil {
						b.Fatal(err)
					}
					b.ResetTimer()

					for i := 0; i < b.N; i++ {
						chs.Pick()
					}
				})

				b.Run("lib=menshen weightedrand", func(b *testing.B) {
					chs, err := NewChooser(wr_menshen_choices...)
					if err != nil {
						b.Fatal(err)
					}
					b.ResetTimer()

					for i := 0; i < b.N; i++ {
						_, _ = chs.PickRandom()
					}
				})
			})
		})
	}
}

func BenchmarkSingle(b *testing.B) {
	if testing.Short() {
		b.Skip()
	}

	for n := BMMinChoices; n <= BMMaxChoices; n *= 10 {
		b.Run(fmt.Sprintf("size=%s", fmt1eN(n)), func(b *testing.B) {
			wr_choices := mockChoices(b, n)
			ru_choices := convertChoices(b, wr_choices)
			wr_menshen_choices := convertChoicesToMenshenWeightedRand(b, wr_choices)

			b.Run("lib=randutil", func(b *testing.B) {
				for i := 0; i < b.N; i++ {
					_, _ = randutil.WeightedChoice(ru_choices)
				}
			})

			b.Run("lib=weightedrand", func(b *testing.B) {
				for i := 0; i < b.N; i++ {
					// never actually do this, this is not how the library is used
					chs, _ := mrothwr.NewChooser(wr_choices...)
					chs.Pick()
				}
			})

			b.Run("lib=menshen weightedrand", func(b *testing.B) {
				for i := 0; i < b.N; i++ {
					// never actually do this, this is not how the library is used
					chs, _ := NewChooser(wr_menshen_choices...)
					_, _ = chs.PickRandom()
				}
			})
		})
	}
}

func mockChoices(tb testing.TB, n int) []mrothwr.Choice[rune, uint] {
	tb.Helper()
	choices := make([]mrothwr.Choice[rune, uint], 0, n)
	for i := 0; i < n; i++ {
		s := '🥑'
		w := rand.Intn(10) + 1
		c := mrothwr.NewChoice(s, uint(w))
		choices = append(choices, c)
	}
	return choices
}

func convertChoices(tb testing.TB, cs []mrothwr.Choice[rune, uint]) []randutil.Choice {
	tb.Helper()
	res := make([]randutil.Choice, len(cs))
	for i, c := range cs {
		res[i] = randutil.Choice{Weight: int(c.Weight), Item: c.Item}
	}
	return res
}

func convertChoicesToMenshenWeightedRand[T rune](tb testing.TB, cs []mrothwr.Choice[rune, uint]) []Choice[rune] {
	tb.Helper()
	res := make([]Choice[rune], len(cs))
	for i, c := range cs {
		res[i] = Choice[rune]{Weight: int(c.Weight), Item: c.Item}

	}
	return res
}

// fmt1eN returns simplified order of magnitude scientific notation for n,
// e.g. "1e2" for 100, "1e7" for 10 million.
func fmt1eN(n int) string {
	return fmt.Sprintf("1e%d", int(math.Log10(float64(n))))
}

func TestCreateWeightedRandomList(t *testing.T) {
	element1 := "element1"
	element2 := "element2"
	element3 := "element3"
	list := []*string{&element1, &element2, &element3}
	expectedElements := []*string{&element1, &element2, &element3}

	element1First := false
	element2First := false
	element3First := false
	for range 200 {
		randomlist, err := CreateWeightedRandomList(list)
		assert.ElementsMatch(t, expectedElements, randomlist)
		assert.Equal(t, nil, err)
		for i, element := range randomlist {
			if *element == element1 {
				if i == 0 {
					element1First = true
				}
			} else if *element == element2 {
				if i == 0 {
					element2First = true
				}
			} else if *element == element3 {
				if i == 0 {
					element3First = true
				}
			}
		}
		if element1First && element2First && element3First {
			break
		}
	}
	assert.Equal(t, element1First, true, "element 1 at least once at first position in randomized list")
	assert.Equal(t, element2First, true, "element 2 at least once at first position in randomized list")
	assert.Equal(t, element3First, true, "element 3 at least once at first position in randomized list")
}

func TestAppendWeightedRandomChoices(t *testing.T) {
	element1 := "element1"
	element2 := "element2"
	element3 := "element3"
	list := []*string{&element1}

	expectedElements := []*string{&element1, &element2, &element3}
	element2Second := false
	element3Second := false
	for range 100 {
		choices := []Choice[string]{{Item: element2, Weight: 1}, {Item: element3, Weight: 1}}
		resultList, err := AppendWeightedRandomChoices(choices, list)
		assert.Equal(t, nil, err)
		assert.ElementsMatch(t, expectedElements, resultList, "all source elements appear in result list")
		assert.Equal(t, element1, *resultList[0], "element 1 always first, randomized elements appended")
		for i, element := range resultList {
			if *element == element2 {
				if i == 1 {
					element2Second = true
				}
			} else if *element == element3 {
				if i == 1 {
					element3Second = true
				}
			}
		}
	}
	assert.Equal(t, element2Second, true, "element 2 at least once at second position in randomized list")
	assert.Equal(t, element3Second, true, "element 3 at least once at second position in randomized list")
}
